/**
* Klasse, die ein Ordner oder eine Datei repräsentiert
*/
function Anyfile(
    filemanager     // Zuständige FileManager-Instanz
) {
    this._filemanager = filemanager;
}

/**
* Daten in JSON umwandeln
* @return Object    JSON, welches Instanz und Inhalt repräsentiert
*/
Anyfile.prototype.toJSON = function() {
    var json = {
        filename: this.filename
    };
    if (this.isDir) {
        json.data = [];
        for (var z = 0; z < this._data.length; ++z)
            json.data.push(this._data[z].toJSON());
    } else {
        json.data = this.data;
        if (this.type)
            json.type = this.type;
    }
    return json;
}

/**
* Instanz (und ggf. Unterinstanzen) anhand von JSON erstellen
* @return Anyfile   Erstellte Instanz
*/
Anyfile.createFromJSON = function(
    filemanager,    // Zuständige FileManager-Instanz
    json            // JSON, welches zu erstellende Instanz und Inhalt repräsentiert
) {
    var obj = new Anyfile(filemanager);
    obj._filename = json.filename;
    if (json.data instanceof Array) {
        obj.isDir = true;
        for (var z = 0; z < json.data.length; ++z)
            obj.addFile(Anyfile.createFromJSON(filemanager, json.data[z]));
    } else
        obj._data = json.data;
    if (!obj.isDir && json.type)
        obj._type = json.type;
    return obj;
}

// Übergeordneter Ordner
Anyfile.prototype._parentDir = null;
Object.defineProperty(Anyfile.prototype, 'parentDir', {
    get() { return this._parentDir; },
    set(dir) {
        if (((dir instanceof Anyfile) && !dir.isDir) || (this._parentDir == dir))
            return;
        var oldParent = this._parentDir;
        this._parentDir = null;
        if (oldParent instanceof Anyfile)
            oldParent._removeFile(this);
        if (dir instanceof Anyfile) {
            dir.addFile(this);
        } else {
            if (oldParent instanceof Anyfile)
                oldParent.reportChange();
        }
    }
});

// Name der Datei
Anyfile.prototype._filename = 'unbenannt';
Object.defineProperty(Anyfile.prototype, 'filename', {
    get() { return this._filename; },
    set(s) {
        this._filename = s;
        this.reportChange();
    }
});

// Inhalt (Dateiliste oder Textinhalt)
Anyfile.prototype._data = '';
Object.defineProperty(Anyfile.prototype, 'data', {
    get() { return this._data; },
    set(s) {
        this._data = s;
        this.reportChange();
    }
});

// Handelt es sich um einen Ordner
Object.defineProperty(Anyfile.prototype, 'isDir', {
    get() { return (this._data instanceof Array); },
    set(bool) {
        if ((this._data instanceof Array) == bool) return;
        this._data = bool ? [] : '';
        this.reportChange();
    }
});

// MIME-Type der Datei
Anyfile.prototype._type = '';
Object.defineProperty(Anyfile.prototype, 'type', {
    get() { return this._type; },
    set(s) {
        this._type = s;
        this.reportChange();
    }
});

/**
* Änderung an Dateien melden
* @return undefined
*/
Anyfile.prototype.reportChange = function() {
    this._filemanager.reportChange(this);
    /*!*/
}

/**
* Prüft, ob es eine Datei namens filename gibt
* @return Boolean   Gibt es diese Datei?
*/
Anyfile.prototype.hasFile = function(
    filename    // Dateiname der gesuchten Datei
) {
    if (!this.isDir)
        return false;
    for (var z = 0; z < this._data.length; ++z) {
        if (this._data[z].filename == filename)
            return true;
    }
    return false;
}

/**
* Die Datei mit Namen file bzw. file.name aus diesem Ordner entfernen (ohne über die Änderung zu informieren)
* @return undefined
*/
Anyfile.prototype._removeFile = function(
    file        // Dateiname der ggf. zu entfernenden Datei oder Anyfile-Instanz
) {
    if (!this.isDir)
        return;
    if (file instanceof Anyfile)
        file = file.filename;
    this._data = this._data.filter(function(o) { return (o.filename != file); });
}

/**
* Die Datei file in diesen Ordner aufnehmen
* @return undefined
*/
Anyfile.prototype.addFile = function(
    file        // Aufzunehmende Anyfile-Instanz
) {
    if (!this.isDir || !(file instanceof Anyfile))
        return;
    this._removeFile(file);
    this._data.push(file);
    file._parentDir = this;
    this.reportChange();
}

/**
* Pfad bestimmen
* @return String    Pfad zu dieser Datei
*/
Anyfile.prototype.getPath = function() {
    return (this._parentDir ? (this._parentDir.getPath() + '/' + this._filename) : this._filename);
};

/**
* (Unter-)Datei anhand von (relativem) Pfad bestimmen
* @return Anyfile       Zum Pfad gehörende Datei
*/
Anyfile.prototype.getByPath = function(
    path,       // Pfad, zu dem die Datei gesucht werden soll
    noAbsolute  // Optional: Nicht in Erwägung ziehen, dass der Pfad absolut ist
) {
    var arrPath = (path instanceof Array) ? path : path.split('/');
    path = arrPath.slice(0);
    while (arrPath.length && (arrPath[arrPath.length - 1] == ''))
        arrPath.pop();
    if (!arrPath.length)
        return this._filemanager._root;
    var next = arrPath.shift();
    switch (next) {
        case '.':
            next = this;
            break;
        case '..':
            next = this._parentDir;
            break;
        default:
            next = this.getSubFile(next);
            break;
    }
    if (!arrPath.length)
        return next;
    if ((!next || !(next = next.getByPath(arrPath, true))) && !noAbsolute && (path[0] == this._filemanager._root.filename)) {
        path.shift();
        next = this._filemanager._root.getByPath(path, true);
    }
    return next;
}

/**
* Direkte Unterdatei mit Namen filename bestimmen
* @return Anyfile   Gesuchte Unterdatei
*/
Anyfile.prototype.getSubFile = function(
    filename        // Dateiname
) {
    if (!this.isDir)
        return null;
    for (var z = 0; z < this._data.length; ++z) {
        if (this._data[z].filename == filename)
            return this._data[z];
    }
    return null;
}

/**
* Überprüfen, ob this direkt oder indirekt in Anyfile-Instanz dir enthalten ist oder die Instanzen identisch sind
* @return Boolean   Ist this im Ordner dir enthalten?
*/
Anyfile.prototype.isIn = function(
    dir     // Anyfile-Instanz zum fraglichen Ordner
) {
    if (dir == this)
        return true;
    for (var parent = this._parentDir; parent instanceof Anyfile; parent = parent._parentDir) {
        if (parent == dir)
            return true;
    }
    return false;

}

/**
* Überprüfen, ob Anyfile-Instanz file in Ordner this eingefügt werden kann (also Dateiname frei und keine Rekursion)
* @return Boolean       Ergebnis der Prüfung
*/
Anyfile.prototype.canInsert = function(
    file        // Anyfile-Instanz die evtl. eingefügt werden soll
) {
    return ((file instanceof Anyfile) && this.isDir && !this.getSubFile(file.filename));// && !file.isIn(this));
}

/**
* Instanz klonen
* @return Anyfile       Klon
*/
Anyfile.prototype.clone = function(
    boolChildren        // Optional: Auch Kinder mitklonen?
) {
    var clone = new Anyfile(this._filemanager);
    clone._filename = this._filename;
    clone._data = this._data;
    clone._type = this._type;
    if (boolChildren && this.isDir) {
        for (var z = 0; z < this._data.length; ++z)
            this._data[z].clone(true)._parentDir = clone;
    }
    return clone;
}
