/******************************************************************************
Copyright © 2022 Sebastian Roith

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
******************************************************************************/

/**
* Klasse, die Modellierübung verwaltet, die Kommunikation mit der Lernplattform über SCORM-API übernimmt und die Anzeige der nötigen Elemente sicherstellt
*/
function Modellieruebung(
    nodeParent,     // DOM-Knoten, als dessen einzige Kinder die Übung angezeigt werden soll
    arrClassNames,  // Klassen, zwischen denen Beziehungen modelliert werden sollen
    loesung,        // Zu definierende Bezeichnungen; Array aus Objekten der Form { "from": "Klassenname", "to": "Klassenname", "desc": "Beziehungsname" }
    fktVisualize,   // Funktion, die zu einer Lösungseinreichung eine Visualisierung liefert (als DOM-Element)
    beziehungen,    // Optional: Name der Beziehung(en), sofern es sich nicht um die Enthält-Beziehung handelt; als Objekt der Form { "shortName": String, "relationName": String }, ggf. Array
    unwanted,       // Optional: Liste (besonders) unerwünschter Beziehungen mit Rückmeldungen
    task,           // Optional: Arbeitsauftrag ('\n' erlaubt)
    heading,        // Optional: Überschrift für Aufgabe
    countAllowed    // Optional: Anzahl maximal erreichbarer Punkte / Anzahl an Versuchen
) {
    Modellieruebung.parent.call(this);
    nodeParent.innerHTML = '';
    this.msgPlaceGeneral = new ScoTools.MessagePlace(true);
    this.msgPlaceGeneral.appendAsChild(nodeParent);
    this.msgPlace = new ScoTools.MessagePlace();
    this.nodeParent = nodeParent;

    this.arrClassNames = arrClassNames;
    this.loesung = loesung;
    this.fktVisualize = fktVisualize;
    if (unwanted instanceof Array)
        this.unwanted = unwanted;
    if (countAllowed)
        this.countAllowed = countAllowed;

    this.connect();
    this._loadStatus();
    this.establish(beziehungen, task, heading);
    this._getSaved();
    this.updateShortInfo();
    AudioFeedbackTask.cache();
    this.ready = true;
}
Modellieruebung.prototype = Object.create(AudioFeedbackTask.prototype);
Modellieruebung.prototype.constructor = Modellieruebung;
Modellieruebung.parent = AudioFeedbackTask;

// Ist das Spiel gestartet?
Modellieruebung.prototype.ready = false;

// Die Lösung: Zu definierende Bezeichnungen als Array aus Objekten der Form { "from": "Klassenname", "to": "Klassenname", "desc": "Beziehungsname", ["hint": …, "hint_priority": …] }
Modellieruebung.prototype.loesung = [];

// Liste (besonders) unerwünschter Beziehungen mit Rückmeldungen als Array aus Objekten der Form { "from": "Klassenname", "to": "Klassenname", "desc": "Beziehungsname", ["hint": …, "hint_priority": …] }
Modellieruebung.prototype.unwanted = [];

// Funktion, die zu einer Lösungseinreichung eine Visualisierung liefert (als DOM-Element)
Modellieruebung.prototype.fktVisualize = (() => null);

// Klassen, zwischen denen Beziehungen modelliert werden sollen
Modellieruebung.prototype.arrClassNames = [];

// Wie viele Fehlversuche gab es bereits?
Modellieruebung.prototype.countWrong = 0;

// Wie viele Versuche stehen zur Verfügung?
Modellieruebung.prototype.countAllowed = 10;

// Permutationsnummer
Modellieruebung.prototype.permutation = undefined;

// Wurde die Aufgabe bereits gelöst?
Modellieruebung.prototype.alreadySolved = false;

/**
* Werte von LMS lesen und wiederherstellen
* @return undefined
*/
Modellieruebung.prototype._getSaved = function() {
    if (this.isConnected()) {
        var data = this.read('suspendData');
        if (data) {
            data = data.split(',');
            const countWrong = Number.parseInt(data[0]);
            const beziehungen = Number.parseInt(data[1]);
            if (Number.isFinite(countWrong))
                this.countWrong = countWrong;
            if (Number.isFinite(beziehungen))
                this.editor.fromInt(beziehungen);
        }
    }
}

/**
* Lädt vom LMS die zufällige Permutation und wendet diese auf this.arrClassNames an
* @return undefined
*/
Modellieruebung.prototype._loadStatus = function() {
    if (this.isConnected())
        this.permutation = Number.parseInt(this.read('lessonLocation'));
    if (!Number.isFinite(this.permutation))
        this.permutation = Modellieruebung.randomPermutation(this.arrClassNames);
    this.arrClassNames = Modellieruebung.applyPermutation(this.arrClassNames, this.permutation);
    this.alreadySolved = this.getDone();
}

/**
* Auf Änderung der Beziehungen reagieren
* @return undefined
*/
Modellieruebung.prototype._procChange = function() {
    this.doSave();
    if (this._stillMsgWrong) {
        this._stillMsgWrong = false;
        this.updateShortInfo();
    }
}

/**
* Aktuellen Status speichern
* @return undefined
*/
Modellieruebung.prototype.doSave = function(
    boolDone        // Optional: Soll gespeichert werden, dass die Aufgabe gelöst wurde?
) {
    if (boolDone)
        this.alreadySolved = true;
    if (this.getDone()) {
        this.reportTime();
        return;
    }
    if (!this.lmslink.ready)
        return;
    this.save({
        'score': Math.max(0, this.countAllowed - this.countWrong),
        'scoreMin': 0,
        'scoreMax': this.countAllowed,
        'lessonStatus': boolDone ? 'completed' : 'incomplete',
        'lessonLocation': this.permutation,
        'suspendData': this.countWrong + ',' + this.editor.toInt()
    });
}

/**
* Eine zufällige Permutation auswählen
* @return Number        Nummer der Permutation
*/
Modellieruebung.randomPermutation = function(
    values          // Array aus zu permutierenden Werten
) {
    const randInt = ((lowerThan) => Math.floor(Math.random() * lowerThan));
    let factor = 1;
    return values.map((s, index) => (factor *= Math.max(1, index)) * randInt(index + 1)).reduce((a, b) => a + b, 0);
}

/**
* Permutation auf this.arrClassNames anwenden
* @return undefined
*/
Modellieruebung.applyPermutation = function(
    values,         // Array aus zu permutierenden Werten
    permutation     // Permutationsnummer (wie von Modellieruebung.randomPermutation geliefert)
) {
    let indizes = values.map(function(s, index) {
        const rest = permutation % (index + 1);
        permutation = Math.floor(permutation / (index + 1));
        return rest;
    }).reverse();
    for (let i = 0; i < indizes.length; ++i) {
        let part = indizes.slice(0, i);
        while (part.indexOf(indizes[i]) != -1)
            ++indizes[i];
    }
    return indizes.map(v => values[v]);
}

/**
* Bedienelemente für das Lernpaket aufbauen
* @return undefined
*/
Modellieruebung.prototype.establish = function(
    beziehungen,    // Optional: Name der Beziehung(en), sofern es sich nicht um die Enthält-Beziehung handelt; als Objekt der Form { "shortName": String, "relationName": String }, ggf. Array
    task,           // Optional: Arbeitsauftrag ('\n' erlaubt)
    heading         // Optional: Überschrift für Aufgabe
) {
    Modellieruebung.parent.prototype.establish.call(this);
    const container = ScoTools.createHtml('article', null, { 'class': 'tab' });
    this._mainContainer = container;
    ScoTools.createHtml('h2', container, null, heading ? heading : 'Beziehungen modellieren');
    (task ? task : 'Ergänze im Klassendiagramm die fehlenden Beziehungen so, dass Folgendes „gebaut“ werden kann. Nutze dabei so wenig Beziehungen wie möglich:').split('\n').map(
        s => ScoTools.createHtml('p', container, null, s)
    );

    container.appendChild(this.fktVisualize(this.loesung));
    this.msgPlace.appendAsChild(container);
    this.editor = new Beziehungen(container, this.arrClassNames, beziehungen);
    this.editor.onsubmit = (data => this.submit(data));
    this.editor.onchange = (() => this._procChange());
    this.nodeParent.appendChild(container);
}

/**
* Aktuelle Eingabe zur Überprüfung einreichen
* @return undefined
*/
Modellieruebung.prototype.submit = function(
    data        // Optional: Zu überprüfende Eingabe
) {
    const ist = data || this.editor.getData();
    const vergleich = this.editor.diff(this.loesung, this.unwanted);
    if (!vergleich.isEqual && !this.alreadySolved) {
        ++this.countWrong;
        this.doSave();
        this._stillMsgWrong = true;
        this.showMsg((this.countWrong <= this.countAllowed) ? 'Diese fehlerhafte Lösung hat dir einen Punkt gekostet' : 'Diese Lösung ist fehlerhaft', ScoTools.MSGTYPE_FAILED);
    } else {
        this.doSave(true);
        this.updateShortInfo();
    }
    this.renderEvaluation(vergleich);
    this.playAudio(vergleich.isEqual);
    if (vergleich.isEqual)
        this.showFinalMessage();
}

/**
* Text mit Einheit zusammenbauen
* @return String        Menschenlesbare Angabe
*/
Modellieruebung.mitEinheit = function(
    zahl,           // Maßzahl
    einheitViele,   // Einheit, wenn Maßzahl nicht Eins ist
    wennEins        // Rückgabewert, wenn Maßzahl gleich Eins it
) {
    const zahlen = ['null', 'eins', 'zwei', 'drei', 'vier', 'fünf', 'sechs', 'sieben', 'acht', 'neun', 'zehn', 'elf', 'zwölf'];
    if (zahl == 1)
        return wennEins;
    return ((zahlen.length > zahl) ? zahlen[zahl] : zahl) + ' ' + einheitViele;
};

/**
* Status des Lernpakets und noch zu erreichende Punkteanzahl anzeigen
* @return undefined
*/
Modellieruebung.prototype.updateShortInfo = function() {
    const einheitViele = 'von ' + Modellieruebung.mitEinheit(this.countAllowed, 'Punkten', 'nur einem erreichbaren Punkt');
    const wennEins = 'einen ' + einheitViele;

    if (this.alreadySolved) {
        this.showMsg('Aufgabe gelöst und ' + Modellieruebung.mitEinheit(Math.max(0, this.countAllowed - this.countWrong), einheitViele, wennEins) + ' erhalten', ScoTools.MSGTYPE_SUCCESS);
        return;
    }

    let s = 'Aufgabe noch nicht gelöst';
    if (this.countWrong >= this.countAllowed)
        s = 'Nach ' + Modellieruebung.mitEinheit(this.countWrong, 'Fehlversuchen', 'einem Fehlversuch') + ' kannst du keinen Punkt mehr erreichen';
    else {
        s = 'Für die richtige Antwort erhältst du ';
        if (this.countWrong)
            s += 'noch ';
        s += Modellieruebung.mitEinheit(this.countAllowed - this.countWrong, einheitViele, wennEins)
    }
    this.showMsg(s, ScoTools.MSGTYPE_INFO);
}

/**
* Eine Meldung über das Ergebnis der Antwortüberprüfung anzeigen
* @return undefined
*/
Modellieruebung.prototype.renderEvaluation = function(
    vergleich       // Daten der Auswertung
) {
    const container = ScoTools.createHtml('article', null, { 'class': 'tab', 'id': 'auswertung' });
    createHtml('h2', container, null, 'Auswertung');
    const msgs = new ScoTools.MessagePlace(true);
    msgs.appendAsChild(container);
    if (vergleich.isEqual) {
        msgs.add('Dein Klassendiagramm ist pefekt', ScoTools.MSGTYPE_SUCCESS);
        const einheitViele = 'von ' + Modellieruebung.mitEinheit(this.countAllowed, 'Punkten', 'nur einem erreichbaren Punkt');
        const wennEins = 'einen ' + einheitViele;
        const message = Modellieruebung.mitEinheit(Math.max(0, this.countAllowed - this.countWrong), einheitViele, wennEins) + ' erreicht';

        msgs.add(message.charAt(0).toUpperCase() + message.substring(1), ScoTools.MSGTYPE_INFO);
    } else {
        msgs.add('Dein Klassendiagramm ist fehlerhaft', ScoTools.MSGTYPE_FAILED);
        let schlimmsterFehler = null;
        let schlimmstesFehlt = undefined;
        vergleich.missing.forEach(beziehung => {
            if (!schlimmsterFehler || ((Number.isFinite(schlimmsterFehler.hint_priority) ? schlimmsterFehler.hint_priority : Infinity) > (Number.isFinite(beziehung.hint_priority) ? beziehung.hint_priority : Infinity))) {
                schlimmsterFehler = beziehung;
                schlimmstesFehlt = true;
            }
        });
        vergleich.unwanted.forEach(beziehung => {
            if (!schlimmsterFehler || ((Number.isFinite(schlimmsterFehler.hint_priority) ? schlimmsterFehler.hint_priority : Infinity) > (Number.isFinite(beziehung.hint_priority) ? beziehung.hint_priority : Infinity))) {
                schlimmsterFehler = beziehung;
                schlimmstesFehlt = false;
            }
        });

        if (schlimmsterFehler) {
            if (schlimmsterFehler.hint)
                msgs.add(schlimmsterFehler.hint, ScoTools.MSGTYPE_INFO);
            else if (schlimmstesFehlt)
                msgs.add('Die Beziehung „' + schlimmsterFehler.from + ' ' + schlimmsterFehler.desc + ' ' + schlimmsterFehler.to + '“ fehlt', ScoTools.MSGTYPE_INFO);
            else
                msgs.add('Die Beziehung „' + schlimmsterFehler.from + ' ' + schlimmsterFehler.desc + ' ' + schlimmsterFehler.to + '“ ergibt keinen Sinn', ScoTools.MSGTYPE_INFO);
            switch (vergleich.missing.length + vergleich.unwanted.length) {
                case 1:
                    break;
                case 2:
                    msgs.add('Es gibt noch ein weiteres Problem', ScoTools.MSGTYPE_INFO);
                    break;
                default:
                    msgs.add('Es gibt noch weitere Probleme', ScoTools.MSGTYPE_INFO);
            }
        }
    }

    createHtml('h3', container, null, 'Erstelltes Klassendiagramm');
    const oldUrl = this._lastUrl;
    const imgBuilder = new KlassendiagrammSvgBuilder(this.arrClassNames);
    this.editor.getData().forEach(o => { if (o instanceof Object) imgBuilder.addRelation(this.arrClassNames.indexOf(o.from), this.arrClassNames.indexOf(o.to), o.desc); });
    this._lastUrl = imgBuilder.createObjectURL();
    createHtml('img', createHtml('figure', container), { src: this._lastUrl, alt: 'Erstelltes Klassendiagramm', style: 'width:100%;height:auto;max-height:70vh' });

    createHtml('h3', container, null, 'Ist-Zustand entsprechend des Klassendiagramms');
    container.appendChild(this.fktVisualize(this.editor.getData()));

    createHtml('h3', container, null, 'Gewünschter Soll-Zustand');
    container.appendChild(this.fktVisualize(this.loesung));

    createHtml('button', createHtml('div', container, { 'class': 'beziehungseingabe', style: 'margin-top:1em' }), { type: 'button' }, 'Erneut versuchen').addEventListener('click', () => { location.hash = '' });

    if (this._lastEvaluation)
        this._lastEvaluation.parentNode.replaceChild(container, this._lastEvaluation);
    else
        this.nodeParent.insertBefore(container, this._mainContainer);
    this._lastEvaluation = container;
    location.hash = '#auswertung';
    container.scrollIntoView({behavior: "smooth", block: "start" });

    if (oldUrl)
        URL.revokeObjectURL(oldUrl);
}



/**
* Eingabefelder zur Konfiguration einer einzelnen Beziehung innerhalb einer Beziehungseditor-Instanz
*/
function Beziehungseingabe(
    editor          // Zuständige Beziehungseditor-Instanz
) {
    Object.defineProperty(this, 'editor', { enumerable: true, writeable: false, value: editor });
    const nodeLi = createHtml('li', editor.nodeRelations, null, undefined, editor.nodeRelations.lastChild);
    const nodeDiv = createHtml('div', nodeLi);
    const nodeClass1 = createHtml('select', nodeDiv);
    editor.arrClassNames.map(classname => createHtml('option', nodeClass1, null, classname));

    if (editor.arrRelationNames.length == 1)
        createHtml(undefined, nodeDiv, null, ' ' + editor.arrRelationNames[0].long + ' ');
    else {
        createHtml(undefined, nodeDiv, null, ' ');
        const nodeWhat = createHtml('select', nodeDiv);
        editor.arrRelationNames.map(({ long }) => createHtml('option', nodeWhat, null, long));
        Object.defineProperty(this, 'selectRelation', { enumerable: true, writeable: false, value: nodeWhat });
        nodeWhat.addEventListener('change', ev => this._procChange());
        nodeWhat.addEventListener('focus',  ev => this._procFocus());
        nodeWhat.addEventListener('blur',   ev => this._procBlur());
        createHtml(undefined, nodeDiv, null, ' ');
    }
    const nodeClass2 = nodeClass1.cloneNode(true);
    nodeDiv.appendChild(nodeClass2);
    createHtml(undefined, nodeDiv, null, ' ');
    const removeBtn = createHtml('button', nodeDiv, { type: 'button', title: 'Diese Beziehung löschen' }, 'X');

    removeBtn.addEventListener('click', ev => this.remove());
    removeBtn.addEventListener('focus',  ev => this._procFocus());
    removeBtn.addEventListener('blur',   ev => this._procBlur());
    nodeClass1.addEventListener('change', ev => this._procChange());
    nodeClass1.addEventListener('focus',  ev => this._procFocus());
    nodeClass1.addEventListener('blur',   ev => this._procBlur());
    nodeClass2.addEventListener('change', ev => this._procChange());
    nodeClass2.addEventListener('focus',  ev => this._procFocus());
    nodeClass2.addEventListener('blur',   ev => this._procBlur());

    Object.defineProperty(this, 'node', { enumerable: true, writeable: false, value: nodeLi });
    Object.defineProperty(this, 'selectStart', { enumerable: true, writeable: false, value: nodeClass1 });
    Object.defineProperty(this, 'selectEnd', { enumerable: true, writeable: false, value: nodeClass2 });
    Object.defineProperty(this, 'removeBtn', { enumerable: true, writeable: false, value: removeBtn });

    editor.registerBeziehungseingabe(this);
}

// Zuständige Beziehungseditor-Instanz
Beziehungseingabe.prototype.editor = null;

// DOM-Knoten für Listenelement, welches Eingabefelder beinhaltet
Beziehungseingabe.prototype.node = null;

// Eingabefeld für Ausgangsklasse der Beziehung
Beziehungseingabe.prototype.selectStart = null;

// Eingabefeld für Zielklasse der Beziehung
Beziehungseingabe.prototype.selectEnd = null;

// Ggf. Eingabefeld für Auswahl der Beziehungsart
Beziehungseingabe.prototype.selectRelation = null;

// Schaltfläche zum Löschen der Beziehung
Beziehungseingabe.prototype.removeBtn = null;

// Wird angezeigt, dass diese Beziehung mehrfach eingestellt wurde?
Object.defineProperty(Beziehungseingabe.prototype, 'showObsolet', { enumerable: true,
    get() { return (this._nodeObsolet) ? true : false; },
    set(v) {
        if ((v ? true : false) == this.showObsolet)
            return;
        if (v)
            this._nodeObsolet = createHtml('p', this.node.firstChild, null, '(Diese Beziehung wurde mehrfach angegeben)');
        else {
            this._nodeObsolet.parentNode.removeChild(this._nodeObsolet);
            delete this._nodeObsolet;
        }
    }
});

/**
* An Editor melden, wenn sich ausgewählte Klasse verändert hat
* @return undefined
*/
Beziehungseingabe.prototype._procChange = function() {
    this.editor.procBeziehungChange(this);
}

/**
* An Editor melden, dass irgendein Eingabefeld für diese Beziehung den Fokus erhalten hat
* @return undefined
*/
Beziehungseingabe.prototype._procFocus = function() {
    this.editor.procBeziehungFocus(this);
}

/**
* An Editor melden, dass irgendein Eingabefeld für diese Beziehung den Fokus abgegeben hat
* @return undefined
*/
Beziehungseingabe.prototype._procBlur = function() {
    this.editor.procBeziehungBlur(this);
}

/**
* Diese Beziehung löschen
* @return undefined
*/
Beziehungseingabe.prototype.remove = function() {
    if (this.node.parentNode)
        this.node.parentNode.removeChild(this.node);
    this.editor.unregisterBeziehungseingabe(this);
}

/**
* Index der ausgewählten Beziehung ermitteln
* @return Number        Index in this.editor.arrRelationNames der ausgewählten Beziehung
*/
Beziehungseingabe.prototype.getRelationIndex = function() {
    return this.selectRelation ? this.selectRelation.selectedIndex : 0;
}

/**
* Informationen zur beschriebenen Beziehung auslesen
* @return Object        Beschreibung der Beziehung in der Form { "from": Index, "to": Index, "type": Index }
*/
Beziehungseingabe.prototype.getData = function() {
    return { "from": this.selectStart.selectedIndex, "to": this.selectEnd.selectedIndex, "type": this.getRelationIndex() };
}

/**
* Beschreibung auf vorgegebene Beziehung ändern
* @return Boolean       Wurde die Konfiguration verändert?
*/
Beziehungseingabe.prototype.setData = function(
    data        // Objekt mit gewünschten Indizes als Eigenschaften from, to und type
) {
    if (!(data instanceof Object))
        return;
    const changed = [
        { select: this.selectStart, value: data.from, length: this.editor.arrClassNames.length },
        { select: this.selectEnd, value: data.to, length: this.editor.arrClassNames.length },
        { select: this.selectRelation, value: data.type, length: this.editor.arrRelationNames.length }
    ].map(function({ select, value, length }) {
        const realValue = Math.max(0, Math.min(length - 1, value));
        if (select && Number.isFinite(realValue) && (realValue != select.selectedIndex)) {
            select.selectedIndex = realValue;
            return true;
        }
        return false;
    }).reduce((a, b) => a || b);
    if (changed)
        this._procChange();
    return changed;
}

/**
* Erstes Eingabefeld fokusieren
* @return undefined
*/
Beziehungseingabe.prototype.focus = function() {
    this.selectStart.focus();
}

/**
* Prüft, ob ein Eingabefeld den Fokus hat
* @return Boolean       Hat ein Eingabefeld den Fokus?
*/
Beziehungseingabe.prototype.isFocused = function() {
    const arr = [this.selectStart, this.selectEnd, this.removeBtn];
    if (this.selectRelation)
        arr.push(this.selectRelation);
    return (arr.indexOf(document.activeElement) != -1);
}



/**
* Ein Editierformular für Beziehungen (nur eine fix vorgegebene Beziehungsart!)
*/
function Beziehungseditor(
    parentNode,         // DOM-Knoten, dessen letztes Kind der Editor sein soll
    arrClassNames,      // Klassennamen
    arrRelationNames    // Optional: Einstellbare Beziehungen als bei nur einer Beziehung String bzw. Objekt der Form { long: String, desc: String } oder Array aus solchen Elementen
) {
    this.arrClassNames = arrClassNames.slice();
    if ((typeof arrRelationNames == 'string') || (arrRelationNames instanceof String))
        this.arrRelationNames = [ { long: arrRelationNames, desc: arrRelationNames } ];
    else if (arrRelationNames instanceof Array)
        this.arrRelationNames = arrRelationNames.map(function(item, index) {
            if ((typeof item == 'string') || (item instanceof String))
                return { long: item, desc: item };
            return { long: item.long ? item.long : ('Beziehung ' + (index + 1)), desc: item.desc ? item.desc : ('Beziehung' + (index + 1)) }
        });
    else if ((arrRelationNames instanceof Object) && arrRelationNames.long && arrRelationNames.desc)
        this.arrRelationNames = [ { long: arrRelationNames.long, desc: arrRelationNames.desc } ];
    else
        this.arrRelationNames = this.arrRelationNames.slice();

    this._arrRelations = [];
    this.buildDOM(parentNode);
}

// Aufzurufende Funktion bei Änderung der Beziehungen
Beziehungseditor.prototype.onchange = null;

// Aufzurufende Funktion bei Änderung der Beziehungen, die gerade bearbeitet wird
Beziehungseditor.prototype.onfocuschange = null;

// Aufzurufende Funktion bei Weitermeldung
Beziehungseditor.prototype.onsubmit = null;

// Klassennamen der verfügbaren Klassen
Beziehungseditor.prototype.arrClassNames = [];

// Informationen zu den einstellbaren Beziehungen
Beziehungseditor.prototype.arrRelationNames = [ { long: 'enthält', desc: 'enthält' } ];

// Enthaltene Beziehungseingabe-Instanzen
Beziehungseditor.prototype._arrRelations = [];

// Aufzählung aller konfigurierbarer Beziehungen
Beziehungseditor.prototype.nodeRelations = null;

// Soll die Aktuallisierung der Obsolet-Anzeige und die Weitermeldung von Änderungen temporär unterbunden werden?
Beziehungseditor.prototype.suppressUpdate = false;

/**
* Editierformular aufbauen und im DOM-Baum einfügen
* @return undefined
*/
Beziehungseditor.prototype.buildDOM = function(
    parentNode          // DOM-Knoten, dessen letztes Kind der Editor sein soll
) {
    const nodeForm = createHtml('form', parentNode, { action: location.href, method: 'post', 'class': 'beziehungseingabe' });
    const nodeUl = createHtml('ul', nodeForm);
    const nodeLastLi = createHtml('li', nodeUl);
    const addBtn = createHtml('button', createHtml('div', nodeLastLi), { type: 'button' }, 'Weitere Beziehung ergänzen');
    const doneBtn = createHtml('button', nodeForm, { type: 'submit' }, 'Lösung überprüfen');

    addBtn.addEventListener('click', ev => this.add());
    nodeForm.addEventListener('submit', ev => { ev.preventDefault(); ev.stopPropagation(); this.submit(); });

    Object.defineProperty(this, 'nodeRelations', { enumerable: true, writeable: false, value: nodeUl });
    /*!*/
}

/**
* Neu erstellte Beziehungseingabe-Instanz mitverwalten
* @return undefined
*/
Beziehungseditor.prototype.registerBeziehungseingabe = function(
    instance        // Mitzuverwaltende Beziehungseingabe-Instanz
) {
    this._arrRelations.push(instance);
    this.procBeziehungChange(instance);
}

/**
* Beziehungseingabe-Instanz aus Liste der zu verwaltenden Beziehungseingaben streichen
* @return undefined
*/
Beziehungseditor.prototype.unregisterBeziehungseingabe = function(
    instance        // Mitzuverwaltende Beziehungseingabe-Instanz
) {
    this._arrRelations = this._arrRelations.filter(o => (o != instance));
    this.procBeziehungChange(instance);
}

/**
* Auf Änderung einer Beziehung reagieren
* @return undefined
*/
Beziehungseditor.prototype.procBeziehungChange = function(
    instance        // Betroffene Beziehungseingabe-Instanz
) {
    if (this.suppressUpdate)
        return;
    this.updateHintObsolete();
    if (this.onchange instanceof Function)
        this.onchange.call(this, this.getData());
}

/**
* Darauf reagieren, wenn ein Eingabefeld den Fokus erhält
* @return undefined
*/
Beziehungseditor.prototype.procBeziehungFocus = function(
    instance        // Betroffene Beziehungseingabe-Instanz
) {
    if (this.suppressUpdate)
        return;
    if (this._blurTimeout) {
        window.clearTimeout(this._blurTimeout.id);
        if (this._blurTimeout.relation == instance)
            return;
    }
    this._notifyFocusChange(instance);
}

/**
* Darauf reagieren, wenn ein Eingabefeld den Fokus verliert
* @return undefined
*/
Beziehungseditor.prototype.procBeziehungBlur = function(
    instance        // Betroffene Beziehungseingabe-Instanz
) {
    if (this.suppressUpdate)
        return;
    this._blurTimeout = { id: window.setTimeout(() => this._notifyFocusChange(), 50), relation: instance };
}

/**
* Wenn nicht direkt nach dem Verlust des Fokuses ein anderes Eingabefeld zur selben Beziehung den Fokus erhält, Neuzeichnen auslösen
* @return undefined
*/
Beziehungseditor.prototype._notifyFocusChange = function(
    instance        // Optional: Beziehungseingabe-Instanz, die aktuell bearbeitet wird
) {
    const relation = (instance instanceof Beziehungseingabe) ? instance.getData() : null;
    this._blurTimeout = null;
    if (this.onfocuschange instanceof Function)
        this.onfocuschange.call(this, relation);
}

/**
* Eingabefelder für weitere Beziehung einfügen
* @return undefined
*/
Beziehungseditor.prototype.add = function(
    nofocus     // Optional: Neue Zeile nicht fokusieren?
) {
    const suppress = this.suppressUpdate;
    this.suppressUpdate = true;
    const relation = new Beziehungseingabe(this);
    if (!nofocus)
        relation.focus();
    this.suppressUpdate = suppress;
    this.procBeziehungChange();
}

/**
* Eingabe zur Auswahl weitergeben
* @return undefined
*/
Beziehungseditor.prototype.submit = function() {
    if (this.onsubmit instanceof Function)
        this.onsubmit.call(this, this.getData());
}

/**
* Alle Beziehungen löschen
* @return undefined
*/
Beziehungseditor.prototype.clear = function() {
    const suppress = this.suppressUpdate;
    this.suppressUpdate = true;
    this._arrRelations.forEach(r => r.remove());
    this.suppressUpdate = suppress;
    this.procBeziehungChange();
}

/**
* Alle eingegebenen Beziehungen auslesen
* @return Array     Eingaben als sortiertes Feld mit Einträgen der Form { "from": Index, "to": Index, "type": String }
*/
Beziehungseditor.prototype.getData = function() {
    return this._arrRelations.map(
        relation => relation.getData()
    ).sort(
        (a, b) => (a.from - b.from) * 2 * this.arrClassNames.length + a.to - b.from
    ).filter(
        (relation, index, array) => ((index == array.length - 1) || (array[index + 1].from != relation.from) || (array[index + 1].to != relation.to) || (array[index + 1].type != relation.type))
    );
}

/**
* Neu festlegen, welche und wie viele Beziehungen eingetragen sind
* @return undefined
*/
Beziehungseditor.prototype.setData = function(
    data        // Array aus Daten (wie von getData geliefert, aber nicht notwendigerweise sortiert; mehrmals gleiche Einträge erlaubt)
) {
    if (!(data instanceof Array))
        return;
    const suppress = this.suppressUpdate;
    this.suppressUpdate = true;
    while (this._arrRelations.length > data.length)
        this._arrRelations[this._arrRelations.length - 1].remove();
    while (this._arrRelations.length < data.length)
        this.add(true);
    this._arrRelations.forEach(
        (o, index) => o.setData(data[index])
    );
    this.suppressUpdate = suppress;
    this.procBeziehungChange();
}

/**
* Anzeige doppelt vorkommender Beziehungen aktuallisieren
* @return undefined
*/
Beziehungseditor.prototype.updateHintObsolete = function() {
    const arr = this._arrRelations.map(function(relation) {
        const infos = relation.getData();
        infos.relations = [ relation ];
        return infos;
    });
    arr.sort(
        (a, b) => ((a.from - b.from) * 2 * this.arrClassNames.length + a.to - b.to) * this.arrRelationNames.length + a.type - b.type
    );
    arr.forEach(function(relation, index, array) {
        if (index && (array[index - 1].from == relation.from) && (array[index - 1].to == relation.to) && (array[index - 1].type == relation.type))
            relation.relations = relation.relations.concat(array[index - 1].relations);
    });
    arr.filter(
        (relation, index, array) => ((index == array.length - 1) || (array[index + 1].from != relation.from) || (array[index + 1].to != relation.to) || (array[index + 1].type != relation.type))
    ).forEach(({relations}) => relations.forEach(relation => (relation.showObsolet = (relations.length > 1))));
}

/**
* Ermitteln, welche Beziehung gerade bearbeitet wird
* @return Beziehungseingabe         Instanz, die aktuell bearbeitet wird
*/
Beziehungseditor.prototype._getFocused = function() {
    return this._arrRelations.filter(o => o.isFocused()).reduce((a, b) => a || b, null);
}

/**
* Ermitteln, welche Beziehung gerade bearbeitet wird
* @return Object        Angabe der Form { "from": Index, "to": Index, "type": String } oder null
*/
Beziehungseditor.prototype.getFocused = function() {
    const focused = this._getFocused();
    return focused && focused.getData();
}



/**
* Klassendiagramm mit bearbeitbaren Beziehungen und bis zu vier Klassen
*/
function Beziehungen(
    parentNode,         // DOM-Knoten, dessen letztes Kind der Editor sein soll
    arrClassNames,      // Klassennamen
    arrRelationNames    // Optional: Einstellbare Beziehungen als bei nur einer Beziehung String bzw. Objekt der Form { long: String, desc: String } oder Array aus solchen Elementen
) {
    Object.defineProperty(this, 'node', { enumerable: true, writeable: false, value: createHtml('div', parentNode) });
    this._img = createHtml('img', createHtml('figure', this.node), {
        alt: 'Klassendiagramm',
        style: 'width:100%;height:auto;max-height:70vh'
    });
    Object.defineProperty(this, 'editor', { enumerable: true, writeable: false, value: new Beziehungseditor(this.node, arrClassNames, arrRelationNames) });
    this.updateImage();
    this.editor.onchange = (data => this.procChange(data));
    this.editor.onfocuschange = (focused => this.procChange(null, focused));
    this.editor.onsubmit = (() => this.procSubmit());
}

// DOM-Knoten, der Klassendiagramm und Bearbeitungsmöglichkeit dafür enthält
Beziehungen.prototype.node = null;

// Editor für Beziehungen
Beziehungen.prototype.editor = null;

// Auszuführende Funktion, wenn Lösung abgegeben wird
Beziehungen.prototype.onsubmit = null;

// Auszuführende Funktion bei inhaltlichen Änderungen
Beziehungen.prototype.onchange = null;

/**
* Klassendiagramm neu erstellen und anzeigen
* @return undefined
*/
Beziehungen.prototype.updateImage = function(
    beziehungen,        // Optional: Array mit Beziehungen, wie von Beziehungseditor.getData geliefert
    highlighted         // Optional: Hervorgehobene Beziehung
) {
    const oldUrl = this._img.src;
    const imgBuilder = new KlassendiagrammSvgBuilder(this.editor.arrClassNames);
    const data = (beziehungen instanceof Array) ? beziehungen : this.editor.getData();
    const active = (highlighted instanceof Object) ? highlighted : this.editor.getFocused();
    data.forEach(o => { if (o instanceof Object) imgBuilder.addRelation(o.from, o.to, this.editor.arrRelationNames[o.type].desc); });
    if (active)
        imgBuilder.addRelation(active.from, active.to, this.editor.arrRelationNames[active.type].desc, true);
    this._img.src = imgBuilder.createObjectURL();
    if (oldUrl)
        URL.revokeObjectURL(oldUrl);
}

/**
* Auf Änderung der Beziehungen reagieren
* @return undefined
*/
Beziehungen.prototype.procChange = function(
    beziehungen         // Optional: Array mit Beziehungen, wie von Beziehungseditor.getData geliefert
) {
    this.updateImage(beziehungen);
    if (this.onchange instanceof Function)
        this.onchange.call(this);
}

/**
* Auf Absenden des Formulars reagieren
* @return undefined
*/
Beziehungen.prototype.procSubmit = function() {
    if (this.onsubmit instanceof Function)
        this.onsubmit.call(this, this.getData());
}

/**
* Interne Darstellung von Beziehungen durch Indizes in Darstellung durch Strings konvertieren
* @return Array     Beziehungen in Form { "from": "Klassenname", "to": "Klassenname", "desc": "Beziehungskurzname", "long": "Beziehungslangname" }
*/
Beziehungen.indizes2strings = function(
    data,               // Interne Beziehungsdarstellung, also Array der Form { "from": Index, "to": Index, "type": String }
    arrClassNames,      // Liste aller Klassennamen
    arrRelationNames    // Liste aller Beziehungen der Form { "desc": "Beziehungskurzname", "long": "Beziehungslangname" }
) {
    return data.map(o => ({
        "from": arrClassNames[o.from],
        "to": arrClassNames[o.to],
        "desc": arrRelationNames[o.type].desc,
        "long": arrRelationNames[o.type].long
    }));
}

/**
* Darstellung von Beziehungen durch Strings konvertieren in interne Darstellung durch Indizes konvertieren
* @return Array     Interne Beziehungsdarstellung, also Array der Form { "from": Index, "to": Index, "type": String }
*/
Beziehungen.strings2indizes = function(
    data,               // Array aus Beziehungen in Form { "from": "Klassenname", "to": "Klassenname", "desc": "Beziehungskurzname", "long": "Beziehungslangname" }
    arrClassNames,      // Liste aller Klassennamen
    arrRelationNames    // Liste aller Beziehungen der Form { "desc": "Beziehungskurzname", "long": "Beziehungslangname" }
) {
    const desc = arrRelationNames.map(o => o.desc);
    return data.map(o => ({
        "from": Math.max(0, arrClassNames.indexOf(o.from)),
        "to": Math.max(0, arrClassNames.indexOf(o.to)),
        "type": Math.max(0, desc.indexOf(o.desc))
    }));
}

/**
* Eingetragene Beziehungen ermitteln
* @return Array     Beziehungen in Form { "from": "Klassenname", "to": Klassenname", "desc": "Beziehungsname", "long": "Beziehungslangname" }
*/
Beziehungen.prototype.getData = function() {
    return Beziehungen.indizes2strings(this.editor.getData(), this.editor.arrClassNames, this.editor.arrRelationNames);
}

/**
* Beziehungen data in eine Zahl umwandeln
* @return Number        Zahl, die Beziehungen repräsentiert
*/
Beziehungen.toInt = function(
    data,           // Beziehungen als Array der Form { "from": Index, "to": Index, "type": Index }
    countClasses    // Anzahl an Klassennamen
) {
    return data.map(({ from, to, type }) => Math.pow(2, (type * countClasses + to) * countClasses + from)).reduce((a, b) => a + b, 0);
}

/**
* Eingetragene Beziehungen in eine Zahl umwandeln
* @return Number        Zahl, die eingetragene Beziehungen repräsentiert
*/
Beziehungen.prototype.toInt = function() {
    return Beziehungen.toInt(this.editor.getData(), this.editor.arrClassNames.length);
}

/**
* Zahl, die Beziehungen repräsentiert, in interne Darstellung von Beziehungen umwandeln
* @return Array         Beziehungen als Array der Form { "from": Index, "to": Index, "type": Index }
*/
Beziehungen.internalFromInt = function(
    value,              // Zahl, die Beziehungen repräsentiert
    countClassNames,    // Anzahl vorhandener Klassennamen
    countRelations      // Anzahl vorhandener Beziehungen
) {
    let mask = 1
    let pos = 0;
    const data = [];
    while (mask <= value) {
        if (mask & value) {
            data.push({
                from: pos % countClassNames,
                to: Math.floor(pos / countClassNames) % countClassNames,
                type: Math.floor(pos / countClassNames / countClassNames) % countRelations
            });
        }
        ++pos;
        mask *= 2;
    }
    return data;
}

/**
* Zahl, die Beziehungen repräsentiert, in Beziehungen umwandeln
* @return Array         Beziehungen der Form { "from": "Klassenname", "to": Klassenname", "desc": "Beziehungsname", "long": "Beziehungslangname" }
*/
Beziehungen.fromInt = function(
    value,          // Zahl, die Beziehungen repräsentiert
    arrClassNames,  // Alle vorhandenen Klassennamen
    arrRelationNames// Alle vorhandenen Beziehungen
) {
    return Beziehungen.indizes2strings(Beziehungen.internalFromInt(value, arrClassNames.length, arrRelationNames.length), arrClassNames, arrRelationNames);
}

/**
* Eingetragene Beziehungen auf die von Zahl value repräsentierte Beziehungen setzen
* @return undefined
*/
Beziehungen.prototype.fromInt = function(
    value           // Zahl, die Beziehungen repräsentiert
) {
    this.editor.setData(Beziehungen.internalFromInt(value, this.editor.arrClassNames.length, this.editor.arrRelationNames.length));
}

/**
* Vergleicht, inwieweit die Beziehungen data aktuell eingetragen sind
* @return Object        Vergleichsergebnis der Form { isEqual: "Entspricht data genau den aktuell eingetragenen Beziehungen?", missing: "Teilarray aus data mit aktuell fehlenden Beziehungen", unwanted: "Array aus im Vergleich zu data unnötigen aktuell eingetragenen Beziehungen" }
*/
Beziehungen.prototype.diff = function(
    data,           // Array aus gewollten Beziehungen der Form { "from": "Klassenname", "to": Klassenname", "desc": "Beziehungsname", "long": "Beziehungslangname", … }
    unwanted        // Optional: Array ungewollter Beziehungen mit Zusatzinfos der Form { "from": "Klassenname", "to": Klassenname", "desc": "Beziehungsname",  … }
) {
    const arrClassNames = this.editor.arrClassNames;
    const arrRelationNames = this.editor.arrRelationNames;
    const soll = Beziehungen.toInt(Beziehungen.strings2indizes(data, arrClassNames, arrRelationNames), arrClassNames.length);
    const ist = this.toInt();
    const vereinigung = (ist | soll);
    const fktBuildMapper = function(original) {
        return function(o) {
            const arr = original.filter(d => (d.from == o.from) && (d.to == o.to) && (d.desc == o.desc));
            return arr.length ? arr[0] : o;
        };
    }
    return {
        isEqual: (ist == soll),
        missing: Beziehungen.fromInt(vereinigung - ist, arrClassNames, arrRelationNames).map(fktBuildMapper(data)),
        unwanted: Beziehungen.fromInt(vereinigung - soll, arrClassNames, arrRelationNames).map(fktBuildMapper((unwanted instanceof Array) ? unwanted : []))
    };
}

/**
* Prüfen, ob eine Beziehungsliste eine Beziehung laut Vorgabe enthält
* @return Boolean       Enthält beziehungen einen Eintrag entsprechend der anderen Parameter?
*/
Beziehungen.has = function(
    beziehungen,    // Array aus Objekten der Form { "from": Klassenname, "to": Klassenname, "desc": "Beziehungsname" }
    from,           // Optional: Vorgabe für Klassenname "from"
    to,             // Optional: Vorgabe für Klassenname "to"
    desc            // Optional: Vorgabe für Beziehungskurzname "desc"
) {
    return beziehungen.filter(b => (((from === undefined) || (b.from == from)) && ((to === undefined) || (b.to == to)) && ((desc === undefined) || (b.desc == desc)))).length ? true : false;
}



/**
* Erstellt eine Vektorgrafik für ein Klassendiagramm mit bis zu vier Klassen und beliebige Enthält-Beziehungen zwischen diesen Klassen
*/
function KlassendiagrammSvgBuilder(
    arrClassNames       // Ein bis vier Klassennamen für das Klassendiagramm
) {
    this.style = this.style.slice();
    this.arrClassNames = arrClassNames.slice();
    this._arrRelations = [];
}

// Breites der Grafik in Millimetern
KlassendiagrammSvgBuilder.prototype.width = 360.03;

// Zusätzlicher Platz nach links und rechts in Millimetern (z. B., damit Linien dicker dargestellt werden können)
KlassendiagrammSvgBuilder.prototype.padding = 0.5;

// Höhe der Grafik in Millimetern
KlassendiagrammSvgBuilder.prototype.height = 0;//180.92;

// CSS-Regeln
KlassendiagrammSvgBuilder.prototype.style = [
    "text {font-family:'Courier New', monospace;font-size:11.289px;font-weight:bold;text-align:center;text-anchor:middle}",
    ".objektkarte text {fill:#a40000}",
    ".objektkarte rect {fill:#eeeeec;stroke:#000;stroke-width:1;stroke-linecap:round;stroke-linejoin:round;stroke-miterlimit:4}",
    ".beziehung text {fill:#2e3436}",
    ".beziehung path {fill:none;stroke:#000;stroke-miterlimit:10;stroke-width:1}",
    ".objektkarte.highlighted text {fill:#fcaf3e}",
    ".objektkarte.highlighted rect {fill:#204a87;stroke:#ce5c00}",
    ".beziehung.highlighted text tspan.highlighted {fill:#ce5c00}",
    ".beziehung.highlighted path {stroke-width:2;stroke:#ce5c00}"
];

// Mögliche Position usw. von Klassenkarten
KlassendiagrammSvgBuilder.prototype.slotsClasscard = [
    { "left": 33.894, "top": 9.6272, "textOffset": 14.1254, "width": 95, "height": 22, "minImageHeight": 41.3 },
    { "left": 231.639, "top": 9.6272, "textOffset": 14.1254, "width": 95, "height": 22, "minImageHeight": 41.3 },
    { "left": 33.894, "top": 102.7409, "textOffset": 14.1254, "width": 95, "height": 22, "minImageHeight": 134.5 },
    { "left": 231.639, "top": 102.7409, "textOffset": 14.1254, "width": 95, "height": 22, "minImageHeight": 134.5 }
];

// Mögliche Beziehungen und wie sie gezeichnet werden sollen
KlassendiagrammSvgBuilder.prototype.slotsRelation = [
    [
        { "d": "m62.191 31.8149v19.92h-61.691v-31.108h33.394", "left": 31.086395, "top": 49.320936, "arrowAfter": true, "minImageHeight": 61.4 },
        { "d": "m128.899 15.8139h102.74", "left": 180.00745, "top": 13.399437, "arrowAfter": true, "minImageHeight": 0 },
        { "d": "m86.209 31.6269v71.113", "rotated": true, "left": 67.424833, "top": -88.51516, "arrowAfter": true, "minImageHeight": 0 },
        { "d": "m121.129 31.6269v30.743h127.89v40.37", "left": 180.00745, "top": 59.956174, "arrowAfter": true, "minImageHeight": 0 }
    ],[
        { "d": "m128.899 25.4409h102.74", "left": 180.39053, "top": 35.60325, "arrowAfter": false, "minImageHeight": 41.3 },
        { "d": "m301.949 31.6269v65h19.986v-65", "rotated": true, "top": -304.449, "arrowAfter": false, "left": 64.6269, "minImageHeight": 103 },
        { "d": "m76.5790 124.7449v40.37h282.95v-149.3h-32.894", "left": 218.18086, "top": 175.27345, "arrowAfter": false, "minImageHeight": 180.92 },
        { "d": "m283.949 31.6269v71.113", "rotated": true, "left": 67.424833, "top": -286.25983, "arrowAfter": true, "minImageHeight": 0 }
    ], [
        { "d": "m76.5790 31.6269v71.113", "rotated": true, "left": 67.80792, "top": -66.311348, "arrowAfter": false, "minImageHeight": 0 },
        { "d": "m86.209 124.7449v30.743h263.7v-130.04h-23.267", "left": 217.79776, "top": 153.06964, "arrowAfter": true, "minImageHeight": 165.2 },
        { "d": "m66.953 124.9249v19.92h-66.454v-31.108h33.394", "left": 33.467636, "top": 142.4344, "arrowAfter": true, "minImageHeight": 154.5 },
        { "d": "m128.899 108.9249h102.74", "left": 180.00745, "top": 106.51291, "arrowAfter": true, "minImageHeight": 0 }
    ], [
        { "d": "m111.509 31.6269v40.37h127.89v30.743", "left": 180.39053, "top": 82.159986, "arrowAfter": false, "minImageHeight": 0 },
        { "d": "m274.329 31.6269v71.113", "rotated": true, "left": 67.80792, "top": -264.056, "arrowAfter": false, "minImageHeight": 0 },
        { "d": "m128.899 118.5549h102.74", "left": 180.39053, "top": 128.71672, "arrowAfter": false, "minImageHeight": 0 },
        { "d": "m274.329 124.9249v19.92h66.454v-31.108h-14.14", "left": 307.29294, "top": 142.4344, "arrowAfter": true, "minImageHeight": 154.5 }
    ]
];

// Beschriftungen für tatsächlich genutzte Klassenkarten
KlassendiagrammSvgBuilder.prototype.arrClassNames = [];

// Tatsächlich genutzte Beziehungen der Form [ { from: Index, to: Index, desc: Beschriftung }, … ]
KlassendiagrammSvgBuilder.prototype._arrRelations = [];

// Hervorgehobene Beziehung
KlassendiagrammSvgBuilder.prototype._highlighted = null;

/**
* Erstellt SVG-Quelltext für die (index + 1)-te Klassenkarte
* @return String    Erstellter SVG-Quelltext (bzw. '', wenn für Klassenkarte keine Positionsinformationen bzw. Klassenname vorhanden ist)
*/
KlassendiagrammSvgBuilder.prototype._getClasscardSvg = function(
    index       // Welche Klassenkarte? Entsprechener Index von this.slotsClasscard
) {
    if ((this.slotsClasscard.length <= index) || (this.arrClassNames.length <= index))
        return '';
    const parts = [];
    const slot = this.slotsClasscard[index];
    this.height = Math.max(this.height, slot.minImageHeight);
    if ((this._highlighted instanceof Object) && ((index == this._highlighted.from) || (index == this._highlighted.to)))
        parts.push('<g class="objektkarte highlighted">');
    else
        parts.push('<g class="objektkarte">');
    parts.push('<rect x="' + slot.left + '" y="' + slot.top + '" width="' + slot.width + '" height="' + slot.height + '" />');
    parts.push('<text x="' + (slot.left + slot.width / 2) + '" y="' + (slot.top + slot.textOffset) + '">');
    parts.push('<tspan>' + this.arrClassNames[index] + '</tspan>');
    parts.push('</text></g>');
    return parts.join('');
}

/**
* Erstellt SVG-Quelltext für eine Beziehung
* @return String    Erstellter SVG-Quelltext (bzw. '', wenn für Klassenkarte keine Positionsinformationen bzw. Klassenname vorhanden ist)
*/
KlassendiagrammSvgBuilder.prototype._getRelationSvg = function(
    relation    // Beziehung in Form { from: Index, to: Index, desc: Array of String }
) {
    const from = relation.from;
    const to = relation.to;
    if (!(this.slotsRelation[from] instanceof Array) || !(this.slotsRelation[from][to] instanceof Object))
        return '';
    const slot = this.slotsRelation[from][to];
    this.height = Math.max(this.height, slot.minImageHeight);
    const parts = [];
    const hasHighlight = ((this._highlighted instanceof Object) && (this._highlighted.from == from) && (this._highlighted.to == to));
    parts.push(hasHighlight ? '<g class="beziehung highlighted">' : '<g class="beziehung">');
    if (slot.d)
        parts.push('<path d="' + slot.d + '"/>');
    if (relation.desc && relation.desc.length) {
        parts.push('<text ');
        if (slot.rotated)
            parts.push('transform="rotate(90)" ');
        parts.push('x="' + slot.left + '" y="' + slot.top + '">');
        if (!slot.arrowAfter)
            parts.push('<tspan' + (hasHighlight ? ' class="highlighted"' : '') + '>&lt;</tspan><tspan style="font-family:sans-serif"> </tspan>');

        parts.push(relation.desc.map(
            desc => ((hasHighlight && (desc == this._highlighted.desc)) ? '<tspan class="highlighted">' : '<tspan>') + desc + '</tspan>'
        ).join(
            '<tspan style="font-family:sans-serif">/</tspan>'
        ));

        if (slot.arrowAfter)
            parts.push('<tspan style="font-family:sans-serif"> </tspan><tspan' + (hasHighlight ? ' class="highlighted"' : '') + '>&gt;</tspan>');
        parts.push('</text>');
    }
    parts.push('</g>');
    return parts.join('');
}

/**
* URL erstellen, die erstellte SVG-Grafik referenziert
* @return String    URL
*/
KlassendiagrammSvgBuilder.prototype.createObjectURL = function() {
    return window.URL.createObjectURL(new Blob([this.getSvg(true)], {type: 'image/svg+xml'}));
}

/**
* SVG-Quelltext für Objektdiagramm zusammenbauen
* @return String    SVG-Quelltext
*/
KlassendiagrammSvgBuilder.prototype.getSvg = function(
    includeXmlDeclaration
) {
    const parts = [];
    parts.push('<style><![CDATA[' + this.style.join('\n') + ']]></style>');
    this.arrClassNames.map((s, index) => parts.push(this._getClasscardSvg(index)));
    this._arrRelations.sort((a, b) => (a.from - b.from) * 2 * this.arrClassNames.length + a.to - b.to);
    this._arrRelations.map(o => parts.push(this._getRelationSvg(o)));
    parts.push('</svg>');
    parts.unshift('<svg version="1.1" width="' + (this.width + 2 * this.padding) + 'mm" height="' + this.height + 'mm" viewBox="' + (-this.padding) + ' 0 ' + (this.width + 2 * this.padding) + ' ' + this.height + '" xmlns="http://www.w3.org/2000/svg">');
    if (includeXmlDeclaration)
        parts.unshift('<?xml version="1.0" encoding="UTF-8"?>\n');
    return parts.join('');
}

/**
* Neue Beziehung fürs Einzeichnen vormerken
* @return this
*/
KlassendiagrammSvgBuilder.prototype.addRelation = function(
    from,       // Entsprechnerder Index von this.slotsClasscard der Start-Klassenkarte der Beziehung
    to,         // Entsprechnerder Index von this.slotsClasscard der End-Klassenkarte der Beziehung
    desc,       // Bezeichnung der Beziehung
    highlighted // Optional: Soll diese Beziehung hervorgehoben werden?
) {
    const vorhanden = this._arrRelations.filter(o => ((o.to == to) && (o.from == from)))[0];
    if (vorhanden) {
        if (vorhanden.desc.indexOf(desc) == -1)
            vorhanden.desc.push(desc);
    } else
        this._arrRelations.push({ from: from, to: to, desc: [ desc ] });
    if (highlighted)
        this._highlighted = { from, to, desc };
    return this;
}
